/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "portLib.h"

/* standard librairies */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>

#include <portLib.h>
#include "server/velodyneHeader.h"

/* Other libs includes */
#include <pomStruct.h>
#include <t3d/v3d.h>
#include <libimages3d.h>
#include <velodyne/velodyneLib.h>
#include <posterLib.h>
#include <velodyneClient.h>
#include "velodyneBuffer.h"

#include <sys/times.h>

/* Returns an angle value in degrees within [0 ; 360] */
double
velodyne_modulo360 (double angle) {
  while (angle >= 360.0) angle -= 360.0;
  while (angle < 0.0) angle += 360.0;
  return (angle);
}

/* Fill the line of the im3d with the points of a cartesian packet */
static inline STATUS
velodyne_fillLineWithPacket (velodyne3DImage*	im3d,
			     VELODYNE_CARTESIAN_PACKET_STR* packet,
			     POM_SENSOR_POS*	currentPos,
			     POM_SENSOR_POS*	startPos,
			     int			numPacket)
{
  int		j;
  velodyne3DPoint	data_point3d;
  T3D		sensor2ToMain2;
  T3D		main2ToOrigin;
  T3D		sensor2ToOrigin;
  T3D		main1ToOrigin;
  T3D		originToMain1;
  T3D		sensor2ToMain1;
  T3D		main1ToSensor1;
  T3D		sensor2ToSensor1;
  /* The suffix 2 stands for 'current', 1 stands for 'start' */

  float x, y, z;
  double*  c; 

  t3dInit(&sensor2ToMain2, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&main2ToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&sensor2ToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&main1ToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&originToMain1, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&sensor2ToMain1, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&main1ToSensor1, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&sensor2ToSensor1, T3D_BRYAN, T3D_ALLOW_CONVERSION);

  sensor2ToMain2.euler.euler[0] = currentPos->sensorToMain.euler.yaw;
  sensor2ToMain2.euler.euler[1] = currentPos->sensorToMain.euler.pitch;
  sensor2ToMain2.euler.euler[2] = currentPos->sensorToMain.euler.roll;
  sensor2ToMain2.euler.euler[3] = currentPos->sensorToMain.euler.x;
  sensor2ToMain2.euler.euler[4] = currentPos->sensorToMain.euler.y;
  sensor2ToMain2.euler.euler[5] = currentPos->sensorToMain.euler.z;

  main2ToOrigin.euler.euler[0] = currentPos->mainToOrigin.euler.yaw;
  main2ToOrigin.euler.euler[1] = currentPos->mainToOrigin.euler.pitch;
  main2ToOrigin.euler.euler[2] = currentPos->mainToOrigin.euler.roll;
  main2ToOrigin.euler.euler[3] = currentPos->mainToOrigin.euler.x;
  main2ToOrigin.euler.euler[4] = currentPos->mainToOrigin.euler.y;
  main2ToOrigin.euler.euler[5] = currentPos->mainToOrigin.euler.z;

  t3dCompIn (&sensor2ToOrigin, &sensor2ToMain2, &main2ToOrigin);

  main1ToOrigin.euler.euler[0] = startPos->mainToOrigin.euler.yaw;
  main1ToOrigin.euler.euler[1] = startPos->mainToOrigin.euler.pitch;
  main1ToOrigin.euler.euler[2] = startPos->mainToOrigin.euler.roll;
  main1ToOrigin.euler.euler[3] = startPos->mainToOrigin.euler.x;
  main1ToOrigin.euler.euler[4] = startPos->mainToOrigin.euler.y;
  main1ToOrigin.euler.euler[5] = startPos->mainToOrigin.euler.z;

  t3dInvertIn(&originToMain1, &main1ToOrigin);
  t3dCompIn (&sensor2ToMain1, &sensor2ToOrigin, &originToMain1);

  t3dInvertIn(&main1ToSensor1, &sensor2ToMain2);  /* Note: sensor2ToMain2 =
						     sensor1ToMain1 */
  t3dCompIn (&sensor2ToSensor1, &sensor2ToMain1, &main1ToSensor1);

  if (!t3dOk(&sensor2ToSensor1)) {
    fprintf (stderr, "velodyne: fillLineWithPacket: bad t3d\n");
    return ERROR;
  }

  if (!t3dConvertTo(T3D_MATRIX, &sensor2ToSensor1)) {
    fprintf (stderr, "velodyne: fillLineWithPacket: can't convert t3d\n");
    return ERROR;
  }

  if (numPacket <= VELODYNE_NB_LINES) {
    for (j = 0; j < VELODYNE_NUM_SHOTS*VELODYNE_NUM_LASERS; j++) {
      if (packet->points[j].status == VELODYNE_BAD_POINT)
	im3d->points[numPacket*VELODYNE_NB_COLUMNS+j].status = VELODYNE_BAD_3DPOINT;
      else {
		  c = sensor2ToSensor1.matrix.matrix;
		  x = packet->points[j].x;
		  y = packet->points[j].y;
		  z = packet->points[j].z;
		  im3d->points[numPacket*VELODYNE_NB_COLUMNS+j].status = VELODYNE_GOOD_3DPOINT;
		  im3d->points[numPacket*VELODYNE_NB_COLUMNS+j].intensity = packet->points[j].intensity;
		  im3d->points[numPacket*VELODYNE_NB_COLUMNS+j].coordinates[0] = 
			x * *c + y * *(c+1) + z * *(c+2) + *(c+3);
		  c+= 4;
		  im3d->points[numPacket*VELODYNE_NB_COLUMNS+j].coordinates[1] = 
			x * *c + y * *(c+1) + z * *(c+2) + *(c+3);
		  c+= 4;
		  im3d->points[numPacket*VELODYNE_NB_COLUMNS+j].coordinates[2] = 
			x * *c + y * *(c+1) + z * *(c+2) + *(c+3);
		  im3d->points[numPacket*VELODYNE_NB_COLUMNS+j].intensity = packet->points[j].intensity;
      }
    }
    im3d->nbLines = numPacket;
  }
  else {
    fprintf (stderr, "velodyne: warning, number of packets exceeding poster size\n");
    return ERROR;
  }
  return OK;
}

/**********************************************************************
 *
 * Extracts one scan from a velodyne stream. Each packet is a set
 * VELODYNE_NUM_SHOT (=12) shots, with angles attached. The angles are encoded
 * in 100th of degrees (range from 0 to 35999).
 *
 * The angle of the first shot from each packet is used to check that a full
 * scan has been acquired (one could have checked the angles of each shot
 * within a packet, but this becomes a bit tedious -- and not really necessary
 * after all).
 *
 **********************************************************************/

ACTIVITY_EVENT
velodyne_grabOneScan(velodyneBuffer*		buffer,
		     velodyneOneShotParams*	oneShotParams,
		     velodyne3DImage*		im3d)

{
  int	j;
  VELODYNE_PACKET_STR* packet;
  VELODYNE_CARTESIAN_PACKET_STR cartesianPacket;
  int	currentHeading;
  int	previousHeading;	/* To check the progression of the heading
				   angle  */
  int	nbPacketsInScan = 0;
  int	deltaHeading = 0;	/* The angle spanned since the fir packet read
				   (encoded in 100th of degrees) */
  POM_SENSOR_POS*	startPos;
  POM_SENSOR_POS*	currentPos;

  /* The following parameters are used to specify which packets must be put in
     the scan */
  double	headingOffset;	/* The heading angle between the sensor and
				   the robot */
  int		firstAngle_velodyne;
  int		lastAngle_velodyne;
  double	firstAngle_1;
  double	lastAngle_1;
  double	firstAngle_2;
  double	lastAngle_2;
  int		reversedAngles = 0;	/* Flag to indicate the first angle is
				   the smallest or not. */
  int		nbPacketsInSector = 0; /* The number of packets in the
					  specified angular sector */

  if (buffer->read == buffer->write)   /* Read the packet only if it is not
					  being updated by the acquisition task */
    goto leavingOut;

  /* Read the first packet, its pos and its heading */
  packet = &(buffer->packets[buffer->read].packet);
  currentPos = &(buffer->packets[buffer->read].position);
  startPos = &(buffer->packets[buffer->read].position);
  currentHeading = packet->data.shots[0].rotational_pos;

  /* Compute the fist/last angle values in the sensor frame. Cf comment in
     velodyne-lib/src/velodyneLib.c re. the Velodyne frame definition, and in
     velodyne-genom/velodyneStruct.h for the interpretation of these angles */
  headingOffset = 180.0 * (currentPos->sensorToMain.euler.yaw) / M_PI;
  /* Computes the angles with respect to the Y axis of the velodyne */
  firstAngle_1 = velodyne_modulo360 (oneShotParams->firstAngle_degrees - headingOffset);
  lastAngle_1 = velodyne_modulo360 (oneShotParams->lastAngle_degrees - headingOffset);
  /* Sets them in the velodyne orientation (clockwise: note that the first and
     second angles are swapped here because of the inversed rotation
     orientation of the Velodyne frame) */
  firstAngle_2 = 360 - lastAngle_1;
  lastAngle_2 = 360 - firstAngle_1; /* (no need to call again modulo360 here) */
  
  /* Converts the angles in hundreth of degrees. */
  firstAngle_velodyne = lround (100 * firstAngle_2);
  lastAngle_velodyne = lround (100 * lastAngle_2 );

  if (firstAngle_velodyne >= lastAngle_velodyne) 
    reversedAngles = 1;		/* (The first/last angles are interpreted
				   differently depending on this variable) */

  /* Put the packet in the line of the im3d */
  if (reversedAngles != 0) {
    if ((currentHeading > firstAngle_velodyne) 
	|| (currentHeading < lastAngle_velodyne)) {
      velodyne_convert_packet (packet, &cartesianPacket);
      if (velodyne_fillLineWithPacket (im3d, &cartesianPacket,
				       currentPos, startPos, nbPacketsInScan) != OK)
	return ETHER;
      nbPacketsInSector++;
    }
  }
  else {
    if ((currentHeading < lastAngle_velodyne) 
	&& (currentHeading > firstAngle_velodyne)) {
      velodyne_convert_packet (packet, &cartesianPacket);
      if (velodyne_fillLineWithPacket (im3d, &cartesianPacket,
				       currentPos, startPos, nbPacketsInScan) != OK)
	return ETHER;
      nbPacketsInSector++;
    }
  }
  buffer->read = (buffer->read + 1) % buffer->nbPackets;
  nbPacketsInScan++;
  
  if (buffer->read == buffer->write) /* Read the packet only if it is not
					being updated by the acquisition task */
    goto leavingOut;

  /* Now loop until a full 360 degrees scan has been read */
  while (deltaHeading <= 36000) {
    if (buffer->read == buffer->write)   /* Read the packet only if it is not
					    being updated by the acquisition task */
      goto leavingOut;

    /* Get the current packet, its pos and heading */
    packet = &(buffer->packets[buffer->read].packet);
    currentPos = &(buffer->packets[buffer->read].position);
    previousHeading = currentHeading;
    currentHeading = packet->data.shots[0].rotational_pos;

    /* Update deltaHeading */
    if (currentHeading < previousHeading)
      /* In this case we crossed the 0 angle between previousHeading and
	 currentHeading */ 
      deltaHeading += (currentHeading + (36000 - previousHeading));
    else
      deltaHeading += (currentHeading - previousHeading);

    /* Put the packet in the line of the im3d */
    if (reversedAngles != 0) {
      if ((currentHeading > firstAngle_velodyne) 
	  || (currentHeading < lastAngle_velodyne)) {
	velodyne_convert_packet (packet, &cartesianPacket);
	if (velodyne_fillLineWithPacket (im3d, &cartesianPacket,
					 currentPos, startPos, nbPacketsInScan) != OK)
	  return ETHER;
	nbPacketsInSector++;
      }
    }
    else {
      if ((currentHeading < lastAngle_velodyne) 
	  && (currentHeading > firstAngle_velodyne)) {
	velodyne_convert_packet (packet, &cartesianPacket);
	if (velodyne_fillLineWithPacket (im3d, &cartesianPacket,
					 currentPos, startPos, nbPacketsInScan) != OK)
	  return ETHER;
	nbPacketsInSector++;
      }
    }
    buffer->read = (buffer->read + 1) % buffer->nbPackets;
    nbPacketsInScan++;
  }

  printf ("Nb packets in the specified sector: %d (out of %d total packets)\n", 
	  nbPacketsInScan, nbPacketsInSector);
  
 leavingOut:
  memcpy (&(im3d->position), startPos, sizeof (POM_SENSOR_POS));

  return END;
}
