/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** velodyneExecTaskCodels.c
 **
 ** Codels called by execution task velodyneExecTask
 **
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <portLib.h>
#include "server/velodyneHeader.h"
#include "velodyneBuffer.h"

#include <pomStruct.h>
#include <libeuler/pomEuler.h>
#include <libimages3d.h>
#include <velodyne/velodyneLib.h>


extern VELODYNE_INTERFACE* device;
extern velodyneReadTask	task;

static POSTER_ID	posPosterId;
static POSTER_ID	sensorPosPosterId;

/* Produced posters */
static velodyne3DImage*		exportedPoster;


/* Function prototypes */
extern ACTIVITY_EVENT velodyne_grabOneScan(velodyneBuffer*, velodyneOneShotParams*, velodyne3DImage*);


/*------------------------------------------------------------------------
 *
 * velodyneExecTaskStart  --  Initialization codel
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneExecTaskStart(int *report)
{
  /* Sets all the request parameters to their default value */

  /* initParams */
  SDI_F->initParams.mode = VELODYNE_DEFAULT_MODE;
  SDI_F->initParams.usePom = VELODYNE_DEFAULT_USE_POM;
  strncpy (SDI_F->initParams.posPoster, VELODYNE_DEFAULT_POS_POSTER,
	   VELODYNE_MAX_STRLEN);
  strncpy (SDI_F->initParams.sensorPosPoster, VELODYNE_DEFAULT_SENSOR_POS_POSTER,
	   VELODYNE_MAX_STRLEN);
  strncpy (SDI_F->initParams.inputFileName, VELODYNE_DEFAULT_INPUT_FILENAME,
	   VELODYNE_MAX_STRLEN);

  /* saveParams */
  strncpy (SDI_F->saveParams.baseName, VELODYNE_DEFAULT_BASENAME,
	   VELODYNE_MAX_STRLEN);
  SDI_F->saveParams.num = VELODYNE_DEFAULT_NUM;

  SDI_F->running = 0;

  /* Private structures */
  exportedPoster = posterAddr (VELODYNE_THREEDIMAGE_POSTER_ID);
  if (exportedPoster == NULL)
    return ERROR;

  return OK;
}

/*------------------------------------------------------------------------
 *
 * velodyneExecTaskEnd  --  Termination codel
 *
 * Description: just closes the driver -- if it has been opened
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneExecTaskEnd(void)
{
  return OK;
}

/*------------------------------------------------------------------------
 * OneShot
 *
 * Description: grabs scans of data from the device
 *
 * Reports:      OK
 *              S_velodyne_DEVICE_NOT_OPEN
 *              S_velodyne_ACQUISITION_NOT_STARTED
 *              S_velodyne_PARTIAL_SCAN_ACQUISITION
 *              S_velodyne_CAN_NOT_UPDATE_POSTER
 */

/* velodyneOneShotStart  -  codel START of OneShot
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
velodyneOneShotStart(velodyneOneShotParams* oneShotParams, int *report)
{
  if (SDI_F->running == 0) {
    fprintf (stderr, "velodyne: the acquisition has not been started\n");
    *report = S_velodyne_ACQUISITION_NOT_STARTED;
    return ETHER;
  }
  if (device == NULL) {
	fprintf (stderr, "velodyne: the device is not open\n");
    *report = S_velodyne_DEVICE_NOT_OPEN;
    return ETHER;
  }
  memset (exportedPoster, 0, sizeof (velodyne3DImage));

  return EXEC;
}

/* velodyneOneShotMain  -  codel EXEC of OneShot
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
velodyneOneShotMain(velodyneOneShotParams* oneShotParams, int *report)
{
  ACTIVITY_EVENT	grabError;
  int			i;
  struct timeval tv_start, tv_end;

  // XXX make sure that needOneShot is accessed in an atomic way 
  /* grab a scan */
  task.needOneShot = true; 
  pthread_mutex_lock(&task.mtx);
  grabError = velodyne_grabOneScan (&task.buffer, oneShotParams, exportedPoster);
  pthread_mutex_unlock(&task.mtx);
  task.needOneShot = false; 
  pthread_cond_signal(&task.cond);

  return grabError;
}

/* velodyneOneShotEnd  -  codel END of OneShot
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
velodyneOneShotEnd(velodyneOneShotParams* oneShotParams, int *report)
{
  if (velodyneThreeDImagevelodyne3DImagePosterWrite
      (VELODYNE_THREEDIMAGE_POSTER_ID,
       exportedPoster) != OK) {
    *report = S_velodyne_CAN_NOT_UPDATE_POSTER;
    return FAIL;
  }
  return ETHER;
}

/*------------------------------------------------------------------------
 * Save
 *
 * Description:
 *
 * Reports:      OK
 *              S_velodyne_CAN_NOT_OPEN_FILE
 *              S_velodyne_CAN_NOT_ALLOCATE_3D_IMAGE
 */

/* velodyneSaveExec  -  codel EXEC of Save
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
velodyneSaveExec(int *report)
{
  int	i, j;
  char	fileName[VELODYNE_MAX_STRLEN];
  int	error;
  DATA_IM3D	im3d;
  DATA_PT3D*	point3d;
  velodyne3DPoint	velodynePoint3d;

  /* Allocates the DATA_IM3D */
  memset(&im3d, 0, sizeof(DATA_IM3D));
  if (init_data_im3d (&im3d, VELODYNE_NB_LINES, VELODYNE_NB_COLUMNS,
		      0, 0, 0) != ERR_IM3D_OK) {
    *report = S_velodyne_CAN_NOT_ALLOCATE_3D_IMAGE;
    return ETHER;
  }

  /* Copy the poster into the im3d */
  for (i = 0; i < VELODYNE_NB_LINES; i++) {
    for (j = 0; j < VELODYNE_NB_COLUMNS; j++) {
      point3d = data_im3d_pt (&im3d, i, j);
      velodynePoint3d = exportedPoster->points[i*VELODYNE_NB_COLUMNS+j];
      if (exportedPoster->points[i*VELODYNE_NB_COLUMNS+j].status == VELODYNE_GOOD_3DPOINT) {
	point3d->state = GOOD_PT3D;
	point3d->coord_1 = velodynePoint3d.coordinates[0];
	point3d->coord_2 = velodynePoint3d.coordinates[1];
	point3d->coord_3 = velodynePoint3d.coordinates[2];
      }
    }
  }

  im3d.header.coordonnees = COORDONNEES_CARTESIENNES;
  im3d.header.capteur = CAPTEUR_LASER1;
  im3d.header.repere = REPERE_CAPTEUR;
  im3d.header.points = POINTS_CALCULES;

  /* Compose the fileName and save the im3d */
  sprintf (fileName, "%s.%04d",
	   SDI_F->saveParams.baseName, SDI_F->saveParams.num);
  error = write_data_im3d_binaire (&im3d, fileName);
  if (error != ERR_IM3D_OK) {
    *report = S_velodyne_CAN_NOT_OPEN_FILE;
    return ETHER;
  }

  /* Compose the fileName and save the pom position */
  sprintf (fileName, "%s.pos.%04d",
	   SDI_F->saveParams.baseName, SDI_F->saveParams.num);
  pomWriteSensorPos(fileName, &(exportedPoster->position));

  SDI_F->saveParams.num += 1;

  /* Free the im3d */
  empty_data_im3d(&im3d);

  return ETHER;
}
