/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** velodyneReadTaskCodels.c
 **
 ** Codels called by execution task velodyneReadTask
 **
 **/

#include <portLib.h>
#include <time.h>

/* standard librairies */
#include <stdio.h>
#include <stdlib.h>

#include "server/velodyneHeader.h"
#include "velodyneBuffer.h"


velodyneReadTask task;
VELODYNE_INTERFACE* device = NULL;
static POSTER_ID	posPosterId;
static POSTER_ID	sensorPosPosterId;

static STATUS
velodyneCreateBuffer ()
{
  task.buffer.nbPackets = VELODYNE_NB_LINES;
  task.buffer.read = 0;
  task.buffer.write = 0;
  task.buffer.packets = malloc (sizeof(velodynePacket) * VELODYNE_NB_LINES);
  if (task.buffer.packets == NULL) {
    fprintf (stderr, "Could not allocate buffer of size %d\n", VELODYNE_NB_LINES);
    return ERROR;
  }
}

static STATUS
velodyneDestroyBuffer ()
{
  if (task.buffer.packets != NULL) {
    free (task.buffer.packets);
    task.buffer.packets = NULL;
  }
  task.buffer.nbPackets = 0;
  task.buffer.read = 0;
  task.buffer.write = 0;
  return OK;
}

/*------------------------------------------------------------------------
 *
 * velodyneReadTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneReadTaskInit(int *report)
{
  memset (&task, 0, sizeof (velodyneReadTask));
  task.buffer.packets = NULL;
  return OK;
}

/*------------------------------------------------------------------------
 *
 * velodyneReadTaskEnd  --  Termination codel
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneReadTaskEnd(void)
{
  velodyneDestroyBuffer ();
  if (device != NULL) {
    velodyne_close(device);
    device = NULL;
  }
  return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description: opens the device -- and get the pom poster address if needed
 *
 * Reports:      OK
 *              S_velodyne_INPUT_FILE_NOT_FOUND
 *              S_velodyne_BAD_INIT_PARAMS
 *              S_velodyne_CAN_NOT_OPEN_DEVICE
 *              S_velodyne_POM_POSTER_NOT_FOUND
 */

/* velodyneInitExec  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
velodyneInitExec(VelodyneInitParams *initParams, int *report)
{
  if (device != NULL) {
    velodyne_close(device);
    device = NULL;
  }

  /* Opens the device */
  switch (SDI_F->initParams.mode) {
  case VELODYNE_FROM_FILE:
    device = velodyne_open (SDI_F->initParams.inputFileName);
    break;
  case VELODYNE_LIVE:
    device = velodyne_open ("udp");
    break;
  default:			/* This should actually never occur */
    *report = S_velodyne_BAD_INIT_PARAMS;
    return ETHER;
    break;
  }

  if (device == NULL) {
    fprintf(stderr, "velodyne: can not open the device\n");
    *report = S_velodyne_CAN_NOT_OPEN_DEVICE;
    return ETHER;
  }

  /* Look up for POM position and sensor position posters */
  if (SDI_F->initParams.usePom == VELODYNE_TRUE) {
    if (velodynePOM_POSPosterFind(initParams->posPoster, &posPosterId) == ERROR) {
      fprintf(stderr, "velodyne: cannot find pom position poster\n");
      *report = S_velodyne_POM_POSTER_NOT_FOUND;
      return ETHER;
    }
    if (velodynePOM_SENSOR_POSPosterFind (initParams->sensorPosPoster,
					  &sensorPosPosterId)
	== ERROR) {
      fprintf(stderr, "velodyne: cannot find pom sensor position poster\n");
      *report = S_velodyne_POM_POSTER_NOT_FOUND;
      return ETHER;
    }
  }

  return ETHER;
}

/*
 * XXX Unlocked access to SDI_F->running, but we can guess it is atomic
 * XXX No way to return a useful error to system as the thread run in
 * background
 */
void*
velodyneReadMain (void *report)
{
  static POM_SENSOR_POS pos;
  static int i = 0;

  while (SDI_F->running != 0) {
	if (task.needOneShot) {
		pthread_mutex_lock(&task.mtx);
		pthread_cond_wait(&task.cond, &task.mtx);
	} else {
		pthread_mutex_lock(&task.mtx);
	}

    if (velodyne_read_packet(device, &(task.buffer.packets[task.buffer.write].packet))
	!= VELODYNE_MSG_BUFFER_SIZE) {
      fprintf (stderr, "velodyne: could not read udp packet\n");
      SDI_F->running = 0;
	  pthread_mutex_unlock(&task.mtx);
      return NULL;
    }
    if (SDI_F->initParams.usePom == VELODYNE_TRUE) {
		if (i++ % 50 == 0) {
			if (velodynePOM_SENSOR_POSPosterRead(sensorPosPosterId, &pos)== ERROR) 
			{
				fprintf(stderr, "velodyne: unable to read pom sensorPos poster\n");
				SDI_F->running = 0;
				pthread_mutex_unlock(&task.mtx);
				return NULL;
			}
      } 
	  task.buffer.packets[task.buffer.write].position = pos;
    }

	task.buffer.write = (task.buffer.write + 1) % task.buffer.nbPackets;
	if (task.buffer.write == task.buffer.read)
		task.buffer.read =  (task.buffer.read + 1) % task.buffer.nbPackets;
    pthread_mutex_unlock(&task.mtx);
  }

  return NULL;
}

/*------------------------------------------------------------------------
 * StartAcquisition
 *
 * Description: continously grab the data and fill the data buffer
 *
 * Reports:      OK
 *              S_velodyne_DEVICE_NOT_OPEN
 *              S_velodyne_BAD_PACKET_READING
 *              S_velodyne_ACQUISITION_ALREADY_STARTED
 */

/* velodyneStartAcquisitionStart  -  codel START of StartAcquisition
   Returns:  EXEC ETHER  */
ACTIVITY_EVENT
velodyneStartAcquisitionMain (int *report)
{
  if (SDI_F->running  != 0) {
    fprintf (stderr, "velodyne: acquistion already started %d\n", SDI_F->running);
    *report = S_velodyne_ACQUISITION_ALREADY_STARTED;
    return FAIL;
  }

  /* Allocates the buffer */
  velodyneDestroyBuffer ();
  velodyneCreateBuffer ();

  if (device == NULL) {
    fprintf(stderr, "velodyne: device not opened\n");
    *report = S_velodyne_DEVICE_NOT_OPEN;
    return FAIL;
  }

  if (pthread_mutex_init(&task.mtx, NULL )) {
	fprintf(stderr, "velodyne: can't init mutex\n");
    *report = S_velodyne_NOT_ENOUGH_RESSOURCE;
    return FAIL;
  } ;

  task.needOneShot = false;

  if (pthread_cond_init(&task.cond, NULL)) {
	fprintf(stderr, "velodyne: can't init condition variable\n");
	pthread_mutex_destroy(&task.mtx);
    *report = S_velodyne_NOT_ENOUGH_RESSOURCE;
    return FAIL;
  };

  SDI_F->running  = 1;
  if (pthread_create(&task.thr, NULL, velodyneReadMain, NULL)) {
	fprintf(stderr, "velodyne: can't init readMain thread\n");
	pthread_cond_destroy(&task.cond);
	pthread_mutex_destroy(&task.mtx);
    *report = S_velodyne_NOT_ENOUGH_RESSOURCE;
	SDI_F->running = 0;
    return FAIL;

  } ;

  return ETHER;
}
